/*

Quelle: http://mymathlib.webtrellis.net/matrices.html

*/



#include <math.h>                 // required for fabs() and sqrt()
#include <float.h>                // required for DBL_EPSILON
#include <stdlib.h>     // required for malloc() and free()    
#include <string.h>		// memcpy()


//                        Internally Defined Routines 
static void   One_Real_Eigenvalue(double Hrow[], double eigen_real[],
                                  double eigen_imag[], int row, double shift);
static void   Two_Eigenvalues(double *H, double *S, double eigen_real[],
                                 double eigen_imag[], int n, int k, double t);
static void   Update_Row(double *Hrow, double cos, double sin, int n, int k);
static void   Update_Column(double* H, double cos, double sin, int n, int k);
static void   Update_Transformation(double *S, double cos, double sin,
                                                               int n, int k);
static void   Double_QR_Iteration(double *H, double *S, int row, int min_row,
                                         int n, double* shift, int iteration);
static void   Product_and_Sum_of_Shifts(double *H, int n, int max_row,
                    double* shift, double *trace, double *det, int iteration);
static int    Two_Consecutive_Small_Subdiagonal(double* H, int min_row,
                                int max_row, int n, double trace, double det);
static void   Double_QR_Step(double *H, int min_row, int max_row, int min_col,
                                  double trace, double det, double *S, int n);
static void   Complex_Division(double x, double y, double u, double v,
                                                       double* a, double* b);
static void   BackSubstitution(double *H, double eigen_real[],
                                                  double eigen_imag[], int n);
static void   BackSubstitute_Real_Vector(double *H, double eigen_real[],
                 double eigen_imag[], int row,  double zero_tolerance, int n);
static void   BackSubstitute_Complex_Vector(double *H, double eigen_real[],
                 double eigen_imag[], int row,  double zero_tolerance, int n);
static void   Calculate_Eigenvectors(double *H, double *S, double eigen_real[],
                                                  double eigen_imag[], int n);

////////////
void Interchange_Rows(double *A, int row1, int row2, int ncols);
void Interchange_Columns(double *A, int col1, int col2, int nrows, int ncols);
void Multiply_Row_by_Scalar(double *A, double x, int row, int ncols);

//                        Internally Defined Routines 
static int Search_Rows(double *A, int p[], int n);
static int Search_Cols(double *A, int p[], int bottom_center_block, int n);
static double Power_of_2(double z);
void Copy_Vector(double *d, double *s, int n);
void Balance_Matrix(double *A, double d[], int p[], int n);
void Balance_Matrix_Back_Transform(double *x, double d[], int p[],
                                                             int ncols, int n);
void Identity_Matrix(double *A, int n);
static void Hessenberg_Elementary_Transform(double* H, double *S,
                                                            int perm[], int n);
int Hessenberg_Form_Elementary(double *A, double* S, int n);

int Hessenberg_Form_Orthogonal(double *A, double *U, int n);





int Hessenberg_Form_Orthogonal(double *A, double *U, int n)
{
   int i, k, col;
   double *u;
   double *p_row, *psubdiag;
   double *pA, *pU;
   double sss;                             // signed sqrt of sum of squares
   double scale;
   double innerproduct;

         // n x n matrices for which n <= 2 are already in Hessenberg form

   Identity_Matrix(U, n);
   if (n <= 2) return 0;

         // Reserve auxillary storage, if unavailable, return an error

   u = (double *) malloc(n * sizeof(double));
   if (u == NULL) return -1;

           // For each column use a Householder transformation 
           //   to zero all entries below the subdiagonal.

   for (psubdiag = A + n, col = 0; col < (n - 2); psubdiag += (n+1), col++) {

          // Calculate the signed square root of the sum of squares of the
          // elements below the diagonal.

      for (pA = psubdiag, sss = 0.0, i = col + 1; i < n; pA += n, i++)
         sss += *pA * *pA;
      if (sss == 0.0) continue;
      sss = sqrt(sss);
      if ( *psubdiag >= 0.0 ) sss = -sss;

          // Calculate the Householder transformation Q = I - 2uu'/u'u.

      u[col + 1] = *psubdiag - sss;
      *psubdiag = sss; 
      for (pA = psubdiag + n, i = col + 2; i < n; pA += n, i++) {
         u[i] = *pA;
         *pA = 0.0;
      }

                    // Premultiply A by Q
   
      scale = -1.0 / (sss * u[col+1]);
      for (p_row = psubdiag - col, i = col + 1; i < n; i++) {
         pA = A + n * (col + 1) + i;
         for (innerproduct = 0.0, k = col + 1; k < n; pA += n, k++) 
            innerproduct += u[k] * *pA;
         innerproduct *= scale;
         for (pA = p_row + i, k = col + 1; k < n; pA += n, k++) 
            *pA -= u[k] * innerproduct;
      }
        
                    // Postmultiply QA by Q

      for (p_row = A, i = 0; i < n; p_row += n, i++) {
         for (innerproduct = 0.0, k = col + 1; k < n; k++) 
            innerproduct += u[k] * *(p_row + k);
         innerproduct *= scale;
         for (k = col + 1; k < n; k++) 
            *(p_row + k) -= u[k] * innerproduct;
      }
        
                    // Postmultiply U by (I - 2uu')

      for (i = 0, pU = U; i < n; pU += n, i++) {
         for (innerproduct = 0.0, k = col + 1; k < n; k++) 
            innerproduct += u[k] * *(pU + k);
         innerproduct *= scale;
         for (k = col + 1; k < n; k++) 
            *(pU + k) -= u[k] * innerproduct;
      }

   }

   free(u);

   return 0;
}















static void Hessenberg_Elementary_Transform(double *H, double* S, int perm[],
                                                                         int n)
{
   int i, j;
   double *pS, *pH;
   double x;

   Identity_Matrix(S, n);
   for (i = n - 2; i >= 1; i--) {
      pH = H + n * (i + 1);
      pS = S + n * (i + 1);
      for (j = i + 1; j < n; pH += n, pS += n, j++) {
         *(pS + i) = *(pH + i - 1);
         *(pH + i - 1) = 0.0;
      }
      if (perm[i] != i) {
         pS = S + n * i;
         pH = S + n * perm[i];
         for (j = i; j < n; j++) {
            *(pS + j) = *(pH + j);
            *(pH + j) = 0.0;
         }
         *(pH + i) = 1.0;
      }
   }
}



int Hessenberg_Form_Elementary(double *A, double* S, int n)
{
   int i, j, k, col, row;
   int* perm;
   double *p_row, *pS_row;
   double max;
   double s;
   double *pA, *pB, *pC, *pS;

         // n x n matrices for which n <= 2 are already in Hessenberg form

   if (n <= 1) { *S = 1.0; return 0; }
   if (n == 2) { *S++ = 1.0; *S++ = 0.0; *S++ = 1.0; *S = 0.0; return 0; }

                       // Allocate working memory

   perm = (int*) malloc(n * sizeof(int));
   if (perm == NULL) return -1;             // not enough memory

           // For each column use Elementary transformations 
           //   to zero the entries below the subdiagonal.

   p_row = A + n;
   pS_row = S + n;
   for (col = 0; col < (n - 2); p_row += n, pS_row += n, col++) {

          // Find the row in column "col" with maximum magnitude where 
          // row >= col + 1.                
     
      row = col + 1;
      perm[row] = row;
      for (pA = p_row + col, max = 0.0, i = row; i < n; pA += n, i++)
         if (fabs(*pA) > max) { perm[row] = i; max = fabs(*pA); }
      
          // If perm[row] != row, then interchange row "row" and row
          // perm[row] and interchange column "row" and column perm[row].

      if ( perm[row] != row ) {
         Interchange_Rows(A, row, perm[row], n);
         Interchange_Columns(A, row, perm[row], n, n);
      }
         
          // Zero out the components lying below the subdiagonal.

      pA = p_row + n;
      pS = pS_row + n;
      for (i = col + 2; i < n; pA += n, pS += n, i++) {
         s = *(pA + col) / *(p_row + col);
         for (j = 0; j < n; j++) 
            *(pA + j) -= *(p_row + j) * s;
         *(pS + col) = s;
         for (j = 0, pB = A + col + 1, pC = A + i; j < n; pB +=n, pC += n, j++) 
            *pB += s * *pC;
      }
   }
   pA = A + n + n;
   pS = S + n + n;
   for (i = 2; i < n; pA += n, pS += n, i++) Copy_Vector(pA, pS, i - 1);

   Hessenberg_Elementary_Transform(A, S, perm, n);
   
   free(perm);
   return 0;
}










void Identity_Matrix(double *A, int n)
{
   *A = 1.0;
   if ( n <= 1) return;
   *(A+1) = 0.0;
   memcpy(A+2,A+1, sizeof(double) * (n - 1) );
   *(A+n+1) = 1.0;
   if (n == 2) return;
   memcpy(A+n+2, A+1, sizeof(double) * (n * (n - 1) - 2) ); 
}


void Copy_Vector(double *d, double *s, int n)
{
   memcpy(d, s, sizeof(double) * n);
}


void Interchange_Rows(double *A, int row1, int row2, int ncols)
{
   int i;
   double *pA1, *pA2;
   double temp;

   pA1 = A + row1 * ncols;
   pA2 = A + row2 * ncols;
   for (i = 0; i < ncols; i++) {
      temp = *pA1;
      *pA1++ = *pA2;
      *pA2++ = temp;
   }
}

void Interchange_Columns(double *A, int col1, int col2, int nrows, int ncols)
{
   int i;
   double *pA1, *pA2;
   double temp;

   pA1 = A + col1;
   pA2 = A + col2;
   for (i = 0; i < nrows; pA1 += ncols, pA2 += ncols, i++) {
      temp = *pA1;
      *pA1 = *pA2;
      *pA2 = temp;
   }
}

void Multiply_Row_by_Scalar(double *A, double x, int row, int ncols)
{
	int i;
	for (i = 0; i < ncols; i++)
		(A+row*ncols)[i]*=x;

}


void Balance_Matrix(double *A, double d[], int p[], int n)
{
   int i,j,k;
   int top_center_block, bottom_center_block, converged;
   double *pA, *pC, *pTop;
   double col_norm, row_norm;
   double x,y,z;

                  // Initialize the permutation 'matrix' p[].
   
   for (i = 0; i < n; i++) p[i] = i;

                   // Initialize the diagonal 'matrix' d[].
   
   for (i = 0; i < n; i++) d[i] = 1.0;

               // Search rows for all zero off-diagonal elements
   
   bottom_center_block = Search_Rows(A, p, n);
   p[n+1] = bottom_center_block;

             // Search columns for all zero off-diagonal elements
   
   top_center_block = Search_Cols(A, p, bottom_center_block, n);
   p[n] = top_center_block;

                     // Exit if matrix is upper triangular
 
   if (top_center_block >= bottom_center_block) return;

                          // Balance the center block

   converged = 0;   
   while(!converged) {
      converged = 1;
      pTop = A + top_center_block * n;
      pA = pTop;
      for (i = top_center_block; i <= bottom_center_block; i++, pA += n) {
         col_norm = 0.0;
         row_norm = 0.0;
         pC = pTop + i;
         for (j = top_center_block; j <= bottom_center_block; j++, pC += n) {
            if (j == i) continue;
            col_norm += fabs(*pC);
            row_norm += fabs(*(pA + j));
         }
         x = row_norm / col_norm;
         y = Power_of_2(x);
         z = 1.0 / y;
         if ( (col_norm * y + row_norm * z) < 0.95 * (col_norm + row_norm) ) {
            d[i] *= y;
            pC = A + i;
            for (j = 0; j <= bottom_center_block; j++, pC += n) *pC *= y;
            for (j = top_center_block; j < n; j++) *(pA + j) *= z;
            converged = 0;
         }
      }
   }
}



static int Search_Rows(double *A, int p[], int n)
{
   double *pA;
   int i,j;
   int bottom_center_block;

               // Search rows for all zero off-diagonal elements
   
   bottom_center_block = n-1;

   pA = A + bottom_center_block * n;
   for (i = bottom_center_block; i >= 0; i--, pA -= n) {
      for (j = 0; j < i; j++)
         if (*(pA + j) != 0.0) break;
      if (j < i) continue;
      for (j = i+1; j < n; j++)
         if (*(pA + j) != 0.0) break;
      if (j < n) continue;
      Interchange_Rows(A, i, bottom_center_block, n);
      Interchange_Columns(A, i, bottom_center_block, n, n);
      p[bottom_center_block] = i;
      bottom_center_block--;
   }
   return bottom_center_block;
}



static int Search_Cols(double *A, int p[], int bottom_center_block, int n)
{
   double *pA;
   int i,j;
   int top_center_block;

             // Search columns for all zero off-diagonal elements
   
   top_center_block = 0;
   pA = A + 1;
   for (i = 0; i < bottom_center_block; pA = A + ++i) {
      for (j = 0; j < i; j++, pA += n)
         if (*pA != 0.0) break;
      if (j < i) continue;
      pA += n;
      for (j = i+1; j < n; j++, pA += n)
         if (*pA != 0.0) break;
      if (j < n) continue;
      if (i > top_center_block) {
         Interchange_Rows(A, i, top_center_block, n);
         Interchange_Columns(A, i, top_center_block, n, n);
         p[top_center_block] = i;
         i--;
      }
      top_center_block++;
   }
   return top_center_block;
}



static double Power_of_2(double z)
{
   union { double dbl; short  word[4]; } x,y;
   double y2;
   const short exp_mask = 0x7FF0;              // Exponent mask
   const short bias = 0x3FF0;                  // The exponent of 1.0
   short offset;

   x.dbl = z;
   y.dbl = 1.0;
   offset = x.word[3] - bias + 0x0010;
   if (offset >= 0) y.word[3] = ((offset >> 1) & exp_mask) + bias;
   else y.word[3] = bias - (((-offset) >> 1) & exp_mask);
   y2 = y.dbl * y.dbl;
   if ( x.dbl <= ( y2 * 0.5) ) y.word[3] -= 0x0010;
   if ( x.dbl > (2.0 * y2) ) y.word[3] += 0x0010;
   
   return y.dbl;
}



void Balance_Matrix_Back_Transform(double *x, double d[], int p[],
                                                             int ncols, int n)
{
   int i;

   for (i = 0; i < n; i++) Multiply_Row_by_Scalar(x, d[i], i, ncols);
   for (i = p[n] - 1; i >= 0; i--)
      if ( p[i] > i ) Interchange_Rows(x, i, p[i], ncols);
   for (i = p[n+1] + 1; i < n; i++)
      if ( p[i] < i ) Interchange_Rows(x, i, p[i], ncols);
}


int QR_Hessenberg_Matrix( double *H, double *S, double eigen_real[],
                          double eigen_imag[], int n, int max_iteration_count)
{
   int i;
   int row;
   int iteration;
   int found_eigenvalue;
   double shift = 0.0;
   double* pH;

   for ( row = n - 1; row >= 0; row--) {
      found_eigenvalue = 0;
      for (iteration = 1; iteration <= max_iteration_count; iteration++) {

                      // Search for small subdiagonal element

         for (i = row, pH = H + row * n; i > 0; i--, pH -= n)
            if (fabs(*(pH + i - 1 )) <= DBL_EPSILON *
                       ( fabs(*(pH - n + i - 1)) + fabs(*(pH + i)) ) ) break;

                // If the subdiagonal element on row "row" is small, then
                // that row element is an eigenvalue.  If the subdiagonal 
                // element on row "row-1" is small, then the eigenvalues
                // of the 2x2 diagonal block consisting rows "row-1" and
                // "row" are eigenvalues.  Otherwise perform a double QR
                // iteration.
        
         switch(row - i) {
            case 0: // One real eigenvalue
               One_Real_Eigenvalue(pH, eigen_real, eigen_imag, i, shift);
               found_eigenvalue = 1;
               break;
            case 1: // Either two real eigenvalues or a complex pair
               row--;
               Two_Eigenvalues(H, S, eigen_real, eigen_imag, n, row, shift);
               found_eigenvalue = 1;
               break;    
            default:
               Double_QR_Iteration(H, S, i, row, n, &shift, iteration);
         }
         if (found_eigenvalue) break;
      }
      if (iteration > max_iteration_count) return -1;
   }

   BackSubstitution(H, eigen_real, eigen_imag, n);
   Calculate_Eigenvectors(H, S, eigen_real, eigen_imag, n);

   return 0;
}

                                                                            //
static void One_Real_Eigenvalue(double Hrow[], double eigen_real[],
                                   double eigen_imag[], int row, double shift)
{
   Hrow[row] += shift;      
   eigen_real[row] = Hrow[row];
   eigen_imag[row] = 0.0;
}


                                                                            //
static void Two_Eigenvalues(double *H, double* S, double eigen_real[],
                            double eigen_imag[], int n, int row, double shift)
{
   double p, q, x, discriminant, r;
   double cos, sin;
   double *Hrow = H + n * row;
   double *Hnextrow = Hrow + n;
   int nextrow = row + 1;

   p = 0.5 * (Hrow[row] - Hnextrow[nextrow]);
   x = Hrow[nextrow] * Hnextrow[row];
   discriminant = p * p + x;
   Hrow[row] += shift;
   Hnextrow[nextrow] += shift;
   if (discriminant > 0.0) {                 // pair of real roots
      q = sqrt(discriminant);
      if (p < 0.0) q = p - q; else q += p;
      eigen_real[row] = Hnextrow[nextrow] + q;
      eigen_real[nextrow] = Hnextrow[nextrow] - x / q;
      eigen_imag[row] = 0.0;
      eigen_imag[nextrow] = 0.0;
      r = sqrt(Hnextrow[row]*Hnextrow[row] + q * q);
      sin = Hnextrow[row] / r;
      cos = q / r;
      Update_Row(Hrow, cos, sin, n, row);
      Update_Column(H, cos, sin, n, row);
      Update_Transformation(S, cos, sin, n, row);
   }
   else {                             // pair of complex roots
      eigen_real[nextrow] = eigen_real[row] = Hnextrow[nextrow] + p;
      eigen_imag[row] = sqrt(fabs(discriminant));
      eigen_imag[nextrow] = -eigen_imag[row];
   }
}



static void Update_Row(double *Hrow, double cos, double sin, int n, int row)
{
   double x;
   double *Hnextrow = Hrow + n;
   int i;

   for (i = row; i < n; i++) {
      x = Hrow[i];
      Hrow[i] = cos * x + sin * Hnextrow[i];
      Hnextrow[i] = cos * Hnextrow[i] - sin * x;
   }
}


                                                                            //
static void Update_Column(double* H, double cos, double sin, int n, int col)
{
   double x;
   int i;
   int next_col = col + 1;

   for (i = 0; i <= next_col; i++, H += n) {
      x = H[col];
      H[col] = cos * x + sin * H[next_col];
      H[next_col] = cos * H[next_col] - sin * x;
   }
}



static void Update_Transformation(double *S, double cos, double sin,
                                                                 int n, int k)
{
   double x;
   int i;
   int k1 = k + 1;

   for (i = 0; i < n; i++, S += n) {
      x = S[k];
      S[k] = cos * x + sin * S[k1];
      S[k1] = cos * S[k1] - sin * x;
   }
}



static void Double_QR_Iteration(double *H, double *S, int min_row, int max_row,
                                          int n, double* shift, int iteration) 
{
   int k;
   double trace, det;

   Product_and_Sum_of_Shifts(H, n, max_row, shift, &trace, &det, iteration);
   k = Two_Consecutive_Small_Subdiagonal(H, min_row, max_row, n, trace, det);
   Double_QR_Step(H, min_row, max_row, k, trace, det, S, n);
}



static void Product_and_Sum_of_Shifts(double *H, int n, int max_row,
                     double* shift, double *trace, double *det, int iteration) 
{
   double *pH = H + max_row * n;
   double *p_aux;
   int i;
   int min_col = max_row - 1;

   if ( ( iteration % 10 ) == 0 ) {
      *shift += pH[max_row];
      for (i = 0, p_aux = H; i <= max_row; p_aux += n, i++)
         p_aux[i] -= pH[max_row];
      p_aux = pH - n;
      *trace = fabs(pH[min_col]) + fabs(p_aux[min_col - 1]);
      *det = *trace * *trace;
      *trace *= 1.5;
   }
   else {
      p_aux = pH - n;
      *trace = p_aux[min_col] + pH[max_row];
      *det = p_aux[min_col] * pH[max_row] - p_aux[max_row] * pH[min_col];
   }
};



static int Two_Consecutive_Small_Subdiagonal(double* H, int min_row,
                                 int max_row, int n, double trace, double det)
{
   double x, y ,z, s;
   double* pH;
   int i, k;

   for (k = max_row - 2, pH = H + k * n; k >= min_row; pH -= n, k--) {
      x = (pH[k] * ( pH[k] - trace ) + det) / pH[n+k] + pH[k+1];
      y = pH[k] + pH[n+k+1] - trace;
      z = pH[n + n + k + 1];
      s = fabs(x) + fabs(y) + fabs(z);
      x /= s;
      y /= s;
      z /= s;
      if (k == min_row) break;
      if ( (fabs(pH[k-1]) * (fabs(y) + fabs(z)) ) <= 
          DBL_EPSILON * fabs(x) *
             (fabs(pH[k-1-n]) + fabs(pH[k]) + fabs(pH[n + k + 1])) ) break; 
   }
   for (i = k+2, pH = H + i * n; i <= max_row; pH += n, i++) pH[i-2] = 0.0;
   for (i = k+3, pH = H + i * n; i <= max_row; pH += n, i++) pH[i-3] = 0.0;
   return k;
};



static void Double_QR_Step(double *H, int min_row, int max_row, int min_col,
                                   double trace, double det, double *S, int n)
{
   double s, x, y, z;
   double a, b, c;
   double *pH;
   double *tH;
   double *pS;
   int i,j,k;
   int last_test_row_col = max_row - 1;

   k = min_col;
   pH = H + min_col * n;
   a = (pH[k] * ( pH[k] - trace ) + det) / pH[n+k] + pH[k+1];
   b = pH[k] + pH[n+k+1] - trace;
   c = pH[n + n + k + 1];
   s = fabs(a) + fabs(b) + fabs(c);
   a /= s;
   b /= s;
   c /= s;

   for (; k <= last_test_row_col; k++, pH += n) {
      if ( k > min_col ) {
         c = (k == last_test_row_col) ? 0.0 : pH[n + n + k - 1];
         x = fabs(pH[k-1]) + fabs(pH[n + k - 1]) + fabs(c);
         if ( x == 0.0 ) continue;
         a = pH[k - 1] / x;
         b = pH[n + k - 1] / x;
         c /= x;
      }
      s = sqrt( a * a + b * b + c * c );
      if (a < 0.0) s = -s;
      if ( k > min_col ) pH[k-1] = -s * x;
      else if (min_row != min_col) pH[k-1] = -pH[k-1];
      a += s;
      x = a / s;
      y = b / s;
      z = c / s;
      b /= a;
      c /= a;

                          // Update rows k, k+1, k+2
      for (j = k; j < n; j++) {
         a = pH[j] + b * pH[n+j];
         if ( k != last_test_row_col ) {
            a += c * pH[n + n + j];
            pH[n + n + j] -= a * z;
         }
         pH[n + j] -= a * y;
         pH[j] -= a * x;
      }

                             // Update column k+1

      j = k + 3;
      if (j > max_row) j = max_row;
      for (i = 0, tH = H; i <= j; i++, tH += n) {
         a = x * tH[k] + y * tH[k+1];
         if ( k != last_test_row_col ) {
            a += z * tH[k+2];
            tH[k+2] -= a * c;
         }
         tH[k+1] -= a * b;
         tH[k] -= a;           
      }

                        // Update transformation matrix

      for (i = 0, pS = S; i < n; pS += n, i++) {
         a = x * pS[k] + y * pS[k+1];
         if ( k != last_test_row_col ) {
            a += z * pS[k+2];
            pS[k+2] -= a * c;
         }
         pS[k+1] -= a * b;
         pS[k] -= a;
      }
   }; 
} 



static void BackSubstitution(double *H, double eigen_real[],
                                                   double eigen_imag[], int n)
{
   double zero_tolerance;
   double *pH;
   int i, j, row;

                        // Calculate the zero tolerance

   pH = H;
   zero_tolerance = fabs(pH[0]);
   for (pH += n, i = 1; i < n; pH += n, i++)
      for (j = i-1; j < n; j++) zero_tolerance += fabs(pH[j]);
   zero_tolerance *= DBL_EPSILON;

                           // Start Backsubstitution

   for (row = n-1; row >= 0; row--) {
      if (eigen_imag[row] == 0.0) 
         BackSubstitute_Real_Vector(H, eigen_real, eigen_imag, row,
                                                           zero_tolerance, n);
      else if ( eigen_imag[row] < 0.0 ) 
         BackSubstitute_Complex_Vector(H, eigen_real, eigen_imag, row,
                                                           zero_tolerance, n);
   } 
}



static void BackSubstitute_Real_Vector(double *H, double eigen_real[],
                   double eigen_imag[], int row,  double zero_tolerance, int n)
{
   double *pH;
   double *pV;
   double x,y;
   double u[4];
   double v[2];
   int i,j,k;

   k = row;
   pH = H + row * n;
   pH[row] = 1.0;
   for (i = row - 1, pH -= n; i >= 0; i--, pH -= n) {
      u[0] = pH[i] - eigen_real[row];
      v[0] = pH[row];
      pV = H + n * k;
      for (j = k; j < row; j++, pV += n) v[0] += pH[j] * pV[row];
      if ( eigen_imag[i] < 0.0 ) {
         u[3] = u[0];
         v[1] = v[0];
      } else {
         k = i;
         if (eigen_imag[i] == 0.0) {
            if (u[0] != 0.0) pH[row] = - v[0] / u[0];
            else pH[row] = - v[0] / zero_tolerance;
         } else {
            u[1] = pH[i+1];
            u[2] = pH[n+i];
            x = (eigen_real[i] - eigen_real[row]);
            x *= x;
            x += eigen_imag[i] * eigen_imag[i];
            pH[row] = (u[1] * v[1] - u[3] * v[0]) / x; 
            if ( fabs(u[1]) > fabs(u[3]) )
               pH[n+row] = -(v[0] + u[0] * pH[row]) / u[1];
            else 
               pH[n+row] = -(v[1] + u[2] * pH[row]) / u[3];
         }
      }
   }    
}



static void BackSubstitute_Complex_Vector(double *H, double eigen_real[],
                   double eigen_imag[], int row,  double zero_tolerance, int n)
{
   double *pH;
   double *pV;
   double x,y;
   double u[4];
   double v[2];
   double w[2];
   int i,j,k;

   k = row - 1;
   pH = H + n * row;
   if ( fabs(pH[k]) > fabs(pH[row-n]) ) {
      pH[k-n] = - (pH[row] - eigen_real[row]) / pH[k];
      pH[row-n] = -eigen_imag[row] / pH[k];
   }
   else 
      Complex_Division(-pH[row-n], 0.0,
           pH[k-n]-eigen_real[row], eigen_imag[row], &pH[k-n], &pH[row-n]);
   pH[k] = 1.0;
   pH[row] = 0.0;
   for (i = row - 2, pH = H + n * i; i >= 0; pH -= n, i--) {
      u[0] = pH[i] - eigen_real[row];
      w[0] = pH[row];
      w[1] = 0.0;
      pV = H + k * n;
      for (j = k; j < row; j++, pV+=n) {
         w[0] += pH[j] * pV[row - 1];
         w[1] += pH[j] * pV[row];
      }
      if (eigen_imag[i] < 0.0) {
         u[3] = u[0];
         v[0] = w[0];
         v[1] = w[1];
      } else {
         k = i;
         if (eigen_imag[i] == 0.0) {
            Complex_Division(-w[0], -w[1], u[0], eigen_imag[row], &pH[row-1],
                                                                    &pH[row]);
         }
         else {
            u[1] = pH[i+1];
            u[2] = pH[n + i];
            x = eigen_real[i] - eigen_real[row];
            y = 2.0 * x * eigen_imag[row];
            x = x * x + eigen_imag[i] * eigen_imag[i] 
                                          - eigen_imag[row] * eigen_imag[row];
            if ( x == 0.0 && y == 0.0 ) 
              x = zero_tolerance * ( fabs(u[0]) + fabs(u[1]) + fabs(u[2])
                                       + fabs(u[3]) + fabs(eigen_imag[row]) );
            Complex_Division(u[1]*v[0] - u[3] * w[0] + w[1] * eigen_imag[row],
                           u[1] * v[1] - u[3] * w[1] - w[0] * eigen_imag[row],
                                                   x, y, &pH[row-1], &pH[row]);
            if ( fabs(u[1]) > (fabs(u[3]) + fabs(eigen_imag[row])) ) {
               pH[n+row-1] = -w[0] - u[0] * pH[row-1]
                                           + eigen_imag[row] * pH[row] / u[1];
               pH[n+row] = -w[1] - u[0] * pH[row]
                                          - eigen_imag[row] * pH[row-1] / u[1];
            }
            else {
               Complex_Division(-v[0] - u[2] * pH[row-1], -v[1] - u[2]*pH[row],
                              u[3], eigen_imag[row], &pH[n+row-1], &pH[n+row]);
            }
         } 
      }
   }
}


static void Calculate_Eigenvectors(double *H, double *S, double eigen_real[],
                                                   double eigen_imag[], int n)
{
   double* pH;
   double* pS;
   double x,y;
   int i,j,k;

   for (k = n-1; k >= 0; k--) {
      if (eigen_imag[k] < 0.0) {
         for (i = 0, pS = S; i < n; pS += n, i++) {
            x = 0.0;
            y = 0.0;
            for (j = 0, pH = H; j <= k; pH += n, j++) {
               x += pS[j] * pH[k-1];
               y += pS[j] * pH[k];
            }
            pS[k-1] = x;
            pS[k] = y;
         }
      } else if (eigen_imag[k] == 0.0) { 
         for (i = 0, pS = S; i < n; i++, pS += n) {
            x = 0.0;
            for (j = 0, pH = H; j <= k; j++, pH += n)
               x += pS[j] * pH[k];
            pS[k] = x;
        }
      }
   }
}

 

static void Complex_Division(double x, double y, double u, double v,
                                                         double* a, double* b)
{
   double q = u*u + v*v;

   *a = (x * u + y * v) / q;
   *b = (y * u - x * v) / q;
}

